import logging
logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')
logger = logging.getLogger(__name__)
from redis import Redis
from rq.queue import Queue
from rq.registry import FinishedJobRegistry, StartedJobRegistry
import time, sys, os, glob
import subprocess, click, shutil
import sdmreader
from realfast import rtutils
import rtpipe.parsecands as pc

# set up  
conn0 = Redis(db=0)
conn = Redis(db=1)   # db for tracking ids of tail jobs
trackercount = 2000  # number of tracking jobs (one per scan in db=1) to monitor 
sdmwait = 3600    # timeout in seconds from last update of sdm to assume writing is finished
snrmin = 6.0
sdmArchdir = '/home/mchammer/evla/sdm/' #!!! THIS NEEDS TO BE SET BY A CENTRALIZED SETUP/CONFIG FILE. # dummy dir: /home/cbe-master/realfast/fake_archdir
bdfArchdir = '/lustre/evla/wcbe/data/archive/' #!!! THIS NEEDS TO BE SET BY A CENTRALIZED SETUP/CONFIG FILE.
redishost = os.uname()[1]  # assuming we start on redis host
scanind = 0  # available in state dict, but setting here since is stable and faster this way

@click.command()
@click.option('--qname', default='default', help='Name of queue to monitor')
@click.option('--triggered/--all', default=False, help='Triggered recording of scans or save all? (default: all)')
@click.option('--archive', '-a', is_flag=True, help='After search defines goodscans, set this to create new sdm and archive it.')
@click.option('--verbose', '-v', help='More verbose (e.g. debugging) output', is_flag=True)
@click.option('--nrao_controls_archiving', '-N', help='NRAO controls archiving; i.e. we make our own BDF hardlinks to no_archive and do no archiving at the end. If this is selected, mcaf_monitor MUST ALSO be run with this same option set.', is_flag=True)
@click.option('--production', help='Run code in full production mode (otherwise just runs as test)', is_flag=True)
@click.option('--threshold', help='Detection threshold used to trigger scan archiving (if --triggered set).', type=float, default=0.)
@click.option('--bdfdir', help='Directory to look for bdfs.', default='/lustre/evla/wcbe/data/no_archive')
def monitor(qname, triggered, archive, verbose, nrao_controls_archiving, production, threshold, bdfdir):
    """ Blocking loop that prints the jobs currently being tracked in queue 'qname'.
    Can optionally be set to do triggered data recording (archiving).
    """

    if verbose:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)

    if nrao_controls_archiving:
        logger.info('***WARNING: NRAO WILL CONTROL ARCHIVING. Check to make sure that this option is also set in mcaf_monitor!!')
        assert not archive, '*** INPUT ERROR: "archive" and "nrao_controls_archiving" inputs are incompatible!'

    if production:
        logger.info('***WARNING: Running the production version of the code.***')
        if archive:
            logger.info('***WARNING: Will do archiving.***')
    else:
        logger.info('Running test version of the code. Will NOT actually archive but will print messages.')

    logger.debug('Monitoring queue running in verbose/debug mode.')
    logger.info('Monitoring queue %s in %s recording mode...' % (qname, ['all', 'triggered'][triggered]))
    q = Queue(qname, connection=conn0)
    qs = Queue('slow', connection=conn0)

    jobids0 = []
    q0hist = [0]
    q1hist = [0]
    sdmlastwritten = {}; sdmcount = {}
    while 1:
        jobids = conn.scan(cursor=0, count=trackercount)[1]

        # track history of queue sizes. saves/prints if values change.
        q0len = len(q.jobs)
        q1len = len(jobids)
        if (q0len != q0hist[-1]) or (q1len != q1hist[-1]):
            q0hist.append(q0len)  # track latest size
            q1hist.append(q1len)
            q0hist = q0hist[-10:]   # keep most recent 10
            q1hist = q1hist[-10:]
            logger.info('** Queue size history (newest to oldest) **')
            logger.info('Worker queue:\t%s' % q0hist[::-1])
            logger.info('Tail queue:\t%s' % q1hist[::-1])

        if jobids0 != jobids:
            logger.info('Tracking %d jobs' % len(jobids))
            logger.debug('jobids: %s' % str(jobids))
            sys.stdout.flush()
            jobids0 = jobids

        # filter all jobids to those that are finished pipeline jobs. now assumes only RT.pipeline jobs in queue
        badjobs = [jobids[i] for i in range(len(jobids)) if not q.fetch_job(jobids[i])]  # clean up jobids list first
        if badjobs:
            logger.info('Cleaning up jobs in tail queue with no counterpart in working queue.')
            for jobid in badjobs:
                rtutils.removejob(jobid)

        # update dict of sdm last written times whenever new scan appears for that filename
        filenames = [q.fetch_job(jobid).args[0]['filename'] for jobid in jobids]
        for filename in set(filenames):
            count = filenames.count(filename)
            if sdmcount.has_key(filename):
                if count > sdmcount[filename]:  # if new scan added for filename, then update time stamp
                    logger.info('Updated last written time for %s' % filename)
                    sdmlastwritten[filename] = time.time()
                sdmcount[filename] = count
            else:  # if no key exists, this one is new, so set initialize
                logger.info('Initialized last written time for %s' % filename)
                sdmlastwritten[filename] = time.time()
                sdmcount[filename] = count

        finishedjobs = [q.fetch_job(jobid) for jobid in jobids if q.fetch_job(jobid).is_finished] # and ('RT.pipeline' in q.fetch_job(jobid).func_name)]
        
        # iterate over list of finished tail jobs (one expected per scan)
        for job in finishedjobs:
            d, segments = job.args
            logger.info('Job %s finished with filename %s, scan %s, segments %s' % (str(job.id), d['filename'], d['scan'], str(segments)))
            readytoarchive = True # default assumption is that this file is ready to move to archive and clear from tracking queue

            jobids = conn.scan(cursor=0, count=trackercount)[1]  # refresh jobids to get latest scans_in_queue
            scans_in_queue = [q.fetch_job(jobid).args[0]['scan'] for jobid in jobids if q.fetch_job(jobid).args[0]['filename'] == d['filename']]
            logger.debug("Scans in queue for filename %s: %s" % (d['filename'], scans_in_queue))

            # To be done for each scan:

            # Error check directory usage/settings
            assert 'bunker' not in bdfdir, '*** BDFDIR ERROR: No messing with bunker bdfs!'
            assert 'telcal' not in bdfdir, '*** BDFDIR ERROR: No messing with telcal bdfs!'
            assert 'mchammer' not in d['workdir'] and 'mctest' not in d['workdir'], '*** WORKDIR ERROR: bunker, mchammer, and mctest are off-limits for writing!'
            assert 'mchammer' not in d['filename'] and 'mctest' not in d['filename'], '*** FILENAME ERROR: bunker, mchammer, and mctest are off-limits for writing!'

            # 0) may want to check that other segments finished for this scan. should do so by default ordering in queue

            # 1) merge segments. removes segment pkls, if successfully merged.
            try:
                pc.cleanup(d['workdir'], d['fileroot'], [d['scan']])
            except:
                logger.error('Could not cleanup cands/noise files for fileroot %s and scan %d. Removing from tracking queue.' % (d['fileroot'], d['scan']))
                rtutils.removejob(job.id)
                scans_in_queue.remove(d['scan'])
                continue

            # 2) get metadata (and check that file still available to work with)
            try:
                # Each sc key contains a dictionary. The key is the scan number.                            
                sc,sr = sdmreader.read_metadata(d['filename'], bdfdir=bdfdir)
            except:
                logger.error('Could not parse sdm %s. Removing from tracking queue.' % d['filename'])
                rtutils.removejob(job.id)
                scans_in_queue.remove(d['scan'])
                continue

            # 3) merge candidate and noise files from scans to merge
            # 4) move products into subdirectory "archivedir" and compile notebook
            try:
                if job == finishedjobs[-1]:  # only do summary plot if last in group to keep from getting bogged down with lots of cands

                    # create merge files and notebook products
                    pc.merge_scans(d['workdir'], d['fileroot'], sc.keys())
                    mergepkl = os.path.join(d['workdir'], 'cands_' + d['fileroot'] + '_merge.pkl')
                    noisepkl = os.path.join(d['workdir'], 'noise_' + d['fileroot'] + '_merge.pkl')

                    pc.nbcompile(d['workdir'], d['fileroot'])
                    notebook = os.path.join(d['workdir'], d['fileroot'] + '.ipynb')
                    notebookhtml = os.path.join(d['workdir'], d['fileroot'] + '.html')

                    # push off to archivedir
                    archivestr = '.'.join(d['fileroot'].split('.')[-2:])
                    if not archivestr.replace('.', '').isdigit():  # should have gotten decimal mjd
                        logger.warn('archivestr not parsed correctly ({0}). Will not move final products into subdirectory.'.format(archivestr))
                    else:
                        archivedir = os.path.join(d['workdir'], archivestr)
                        if not os.path.exists(archivedir):
                            os.mkdir(archivedir)
                            logger.info('Creating local archive directory at {0}'.format(archivedir))
                        else:
                            logger.info('Using local archive directory at {0}'.format(archivedir))

                        shutil.copy(mergepkl, archivedir)
                        shutil.copy(noisepkl, archivedir)
                        shutil.copy(notebook, archivedir)
                        shutil.copy(notebookhtml, archivedir)
                        logger.info('Copied merged products into local archive directory {0}'.format(archivedir))
                        # maybe copy GN file in too?
            except:
                logger.info('Trouble merging scans and plotting for scans %s in file %s. Removing from tracking queue.' % (str(sc.keys()), d['fileroot']))
                logger.debug('Exception:\n{0}'.format(sys.exc_info()[0]))
                rtutils.removejob(job.id)
                scans_in_queue.remove(d['scan'])
                continue

            # 5) rsync the interactive html and associated candidate plots out for inspection (note: cand plots may be delayed, so final rsync needed)
            try:
                rtutils.moveplots(d['fileroot'])
            except:
                logger.error('Failed to move cand plots and interactive plot out')

            # 6) if last scan of sdm, start end-of-sb processing. requires all bdf written or sdm not updated in sdmwait period
            allbdfwritten = all([sc[i]['bdfstr'] for i in sc.keys()])
            sdmtimeout = time.time() - sdmlastwritten[d['filename']] > sdmwait
            logger.debug('allbdfwritten = %s. sdmtimeout = %s.' % (str(allbdfwritten), str(sdmtimeout)))
            if (sdmtimeout or allbdfwritten) and (len(scans_in_queue) == 1) and (d['scan'] in scans_in_queue):
                logger.info('This job processed scan %d, the last scan in the queue for %s.' % (d['scan'], d['filename']))

                # 6-2) if doing triggered recording, get scans to save. otherwise, save all scans.
                if triggered:
                    logger.debug('Triggering is on. Saving cal scans and those with candidates.')
                    goodscans = [s for s in sc.keys() if 'CALIBRATE' in sc[s]['intent']]  # minimal set to save

                    # if merged cands available, identify scans to archive.
                    # ultimately, this could be much more clever than finding non-zero count scans.
                    if os.path.exists(mergepkl):
                        goodscans += [sigloc[scanind] for sigloc in pc.thresholdcands(mergepkl, threshold, numberperscan=1)]
                    goodscans = sorted(set(goodscans))  # uniq'd scan list in increasing order
                else:
                    logger.debug('Triggering is off. Saving all scans.')
                    goodscans = sc.keys()

                goodscanstr= ','.join(str(s) for s in goodscans)
                logger.info('Found the following scans to archive: %s' % goodscanstr)

# SLOW QUEUE ON THE SHELF FOR NOW...
#                # 6-3) Edit SDM to remove no-cand scans. Perl script takes SDM work dir, and target directory to place edited SDM.
#                # first determine if this filename is still being worked on by slow queue
#                slowjobids = qs.job_ids # + getstartedjobs('slow')  # working and queued for slow queue
#                remaining = [jobid for jobid in slowjobids if os.path.basename(d['filename']).rstrip('.pkl') in qs.fetch_job(jobid).args[0]]  # these jobs are still open for this file
#
#                if len(remaining) == 0:
#                    logger.info('No jobs for file %s in slow queue. Moving candidate scan data to archive.' % (d['filename']))
#                else:  # slow queue needs more time
#                    logger.info('File %s is still being worked on in slow queue. Will not move to archive yet.' % d['filename'])
#                    logger.debug('remaining jobids: %s' % str(remaining))
#                    readytoarchive = False  # looks like we're not ready! use this below to keep file in tracking queue
#                    continue

                # Once slow queue is finished, either 
                if archive:
                    movetoarchive(d['filename'], d['workdir'].rstrip('/'), goodscanstr, production, bdfdir)
                elif nrao_controls_archiving:
                    logger.debug('Will remove BDFs')
                    removebdfs(d['filename'], d['workdir'].rstrip('/'), production, bdfdir)
                else:
                    logger.debug('Archiving is off.')

                # 6-4) Combine MS files from slow integration into single file. Merges only MS files it finds from provided scan list.
#                try:
#                    rtutils.mergems(d['filename'], sc.keys(), redishost=redishost)
#                except:
#                    logger.info('Failed to merge slow MS files. Continuing...')
 
                # final rsync to get html and cand plots out for inspection
                try:
                    rtutils.moveplots(d['fileroot'])
                except:
                    logger.error('Failed to move cand plots and interactive plot out')

            elif not allbdfwritten and (len(scans_in_queue) == 1) and (d['scan'] in scans_in_queue):
                logger.info('Not all bdf written yet. Keeping last scan of %s in tracking queue.' % d['filename'])
                readytoarchive = False  # looks like we're not ready! use this below to keep file in tracking queue

            else:
                logger.info('Scan %d is not last scan or %s is not finished writing.' % (d['scan'], d['filename']))
                logger.debug('List of bdfstr: %s. scans_in_queue = %s.' % (str([sc[i]['bdfstr'] for i in sc.keys()]), str(scans_in_queue)))

            # job is finished, so remove from db
            if readytoarchive:  # will be false is slow queue not yet empty of relevant jobs
                logger.info('Removing job %s from tracking queue.' % job.id)
                rtutils.removejob(job.id)
                scans_in_queue.remove(d['scan'])
                sys.stdout.flush()
            else:
                logger.info('Keeping job %s from tracking queue.' % job.id)

        sys.stdout.flush()
        time.sleep(1)

def removebdfs(filename, workdir, production, bdfdir):
    """ Moves sdm and bdf associated with filename to archive.
    filename is sdmfile. workdir is place with file.
    goodscanstr is comma-delimited list, which is optional.
    production is boolean for production mode.
    """

    assert filename, 'Need filename to move to archive'

    if not workdir:
        workdir = os.getcwd()
    sc,sr = sdmreader.read_metadata(filename, bdfdir=bdfdir)


    # Clean up all the hardlinks in no_archive for this SB.
    for scan in sc.keys():
        # System safety checks
        assert 'bunker' not in os.path.dirname(sc[scan]['bdfstr']), '*** BDFSTR ERROR: No messing with bunker bdfs!'
        assert 'telcal' not in os.path.dirname(sc[scan]['bdfstr']), '*** BDFSTR ERROR: No messing with telcal bdfs!'
        bdfREMOVE = sc[scan]['bdfstr']
        if bdfREMOVE:
            if not production:
                logger.info('TEST MODE. Would remove BDF %s' % bdfREMOVE.rstrip('/') )
                touch( bdfREMOVE.rstrip('/') + '.delete' )
            else:
                logger.info('Removing BDF %s' % bdfREMOVE.rstrip('/') )
                os.remove( bdfREMOVE.rstrip('/') )


def movetoarchive(filename, workdir, goodscanstr, production, bdfdir):
    """ Moves sdm and bdf associated with filename to archive.
    filename is sdmfile. workdir is place with file.
    goodscanstr is comma-delimited list, which is optional.
    production is boolean for production mode.
    """

    assert filename, 'Need filename to move to archive'

    if not workdir:
        workdir = os.getcwd()
    sc,sr = sdmreader.read_metadata(filename, bdfdir=bdfdir)
    if not goodscanstr:
        goodscanstr = ','.join([str(s) for s in sc.keys() if sc[s]['bdfstr']])
    goodscans = [int(s) for s in goodscanstr.split(',')]

    logger.debug('Archiving is on.')
    logger.debug('Archiving directory info:')
    logger.debug('Workdir: %s' % workdir)    # !! could just be set to os.getcwd()?
    logger.debug('SDMarch: %s' % sdmArchdir)
    logger.debug('SDM:     %s' % filename)
    logger.debug('BDFarch: %s' % bdfArchdir)
    logger.debug('BDFwork: %s' % bdfdir)

    # Set up names of source/target SDM files
    sdmORIG = filename.rstrip('/')
    sdmFROM = filename.rstrip('/') + "_edited"
    sdmTO   = os.path.join(sdmArchdir, os.path.basename(filename.rstrip('/')))

    # safety checks
    if not goodscans:
        raise ValueError, 'No scans found to move to archive (either none provided or none with bdfs).'
    assert 'bunker' not in os.path.dirname(sc[goodscans[0]]['bdfstr']), '*** BDFSTR ERROR: No messing with bunker bdfs!'
    assert 'telcal' not in os.path.dirname(sc[goodscans[0]]['bdfstr']), '*** BDFSTR ERROR: No messing with telcal bdfs!'
    assert not os.path.exists(sdmFROM), 'Edited SDM already exists at %s' % sdmFROM

    if production:
        subprocess.call(['sdm_chop-n-serve.pl', filename, workdir, goodscanstr])   # would be nice to make this Python

    # 4) copy new SDM and good BDFs to archive locations
                    
    # Archive edited SDM
    if not production:
        logger.info('TEST MODE. Would archive SDM %s to %s' % ( sdmFROM, sdmTO ))
        touch(sdmFROM + ".archived")
    else:
        logger.info('Archiving SDM %s to %s' % ( sdmFROM, sdmTO ))
        rtutils.rsync( sdmFROM, sdmTO )

    # Remove old SDM and old edited copy
    if not production:
        logger.info('TEST MODE. Would delete edited SDM %s' % sdmFROM )
        logger.info('TEST MODE. Would delete original SDM %s' % sdmORIG )
        touch(sdmFROM + ".delete")
        touch(sdmORIG + ".delete")
    else: 
        logger.debug('Deleting edited SDM %s' % sdmFROM )
        shutil.rmtree( sdmFROM )
        logger.info('***NOTE (%s): not deleting unedited SDM files yet' % sdmORIG )
        #!!!logger.debug('Deleting original SDM %s' % sdmORIG ) #!!! WHEN CASEY SAYS GO
        #!!!shutil.rmtree( sdmORIG ) #!!! PUT THIS LINE IN WHEN CASEY SAYS GO


    # Archive the BDF (via hardlink to archdir) only if we are controlling the archiving.
    for scan in goodscans:
        bdfFROM = sc[scan]['bdfstr']
        bdfTO   = os.path.join(bdfArchdir, os.path.basename(bdfFROM))
        if not production:
            logger.info('TEST MODE. Would hardlink %s to %s' % ( bdfFROM, bdfTO ))
            touch( bdfFROM + ".archived" )
        else:
            logger.debug('Hardlinking %s to %s' % ( bdfFROM, bdfTO ))
            os.link( bdfFROM, bdfTO )

    # Now delete all the hardlinks in our BDF working directory for this SB.
    # Could also use removebdfs() here.
    for scan in sc.keys():
        bdfREMOVE = sc[scan]['bdfstr']
        if bdfREMOVE:
            if not production:
                logger.info('TEST MODE. Would remove BDF %s' % bdfREMOVE.rstrip('/') )
                touch( bdfREMOVE.rstrip('/') + '.delete' )
            else:
                logger.debug('Removing BDF %s' % bdfREMOVE.rstrip('/') )
                os.remove( bdfREMOVE.rstrip('/') )

def getfinishedjobs(qname='default'):
    """ Get list of job ids in finished registry.
    """

    q = Queue(qname, connection=conn0)
    return FinishedJobRegistry(name=q.name, connection=conn0).get_job_ids()

def getstartedjobs(qname='default'):
    """ Get list of job ids in started registry.
    """

    q = Queue(qname, connection=conn0)
    return StartedJobRegistry(name=q.name, connection=conn0).get_job_ids()

# Temporary method for creating an empty file.
def touch(path):
    with open(path, 'a'):
        os.utime(path, None)




from jinja2 import Environment

# stuff for jinja2 template for index to browse cand summary plots

HTML = """
<html>
<body>

 <ul>
 {% for item in contents %}
  <li><a href="{{ item[1] }}" target="frame">{{ item[0] }}</a> ({{ item[2] }} MB)</li>
 {% endfor %}
 </ul>

</body>
</html>
"""

def gethtmlcontents(directory):
    filenames = glob.glob(os.path.join(directory, 'cands*_merge.html'))
    mjds = ['.'.join(ff.rstrip('_merge.html').split('.')[-2:])
            for ff in filenames]
    sizes = [os.stat(ff).st_size/(1024**2)
             for ff in filenames]
    contents = zip(mjds, [os.path.basename(ff) for ff in filenames], sizes)
    return sorted(contents, key=lambda ff: ff[0])


def writehtml(directory, outname='contents.html'):
    contents = gethtmlcontents(directory)
    contentshtml = Environment().from_string(HTML).render(contents=contents, title='List of Contents')

    with open(os.path.join(directory, outname), 'w') as f:
        f.write(contentshtml)



#! /usr/bin/env python

# Main controller. Sarah Burke Spolaor June 2015
#
# Based on frb_trigger_controller.py by P. Demorest, 2015/02
#
# Listen for OBS packets having a certain 'triggered archiving'
# intent, and perform some as-yet-unspecified action when these
# are recieved.
#

"""
This branch of the code is testing the addition of an option that
archives data as soon as it comes into the CBE, rather than at the end
in queue_monitor.

Need to take care of:

 - Will need no_archive to be functioning so scripts need to include
   this option.

 - Do NOT do any hardlinking at the end in q_m -- otherwise there will
   be double copies pushed into the archive.

 - Delete all BDFs in no_archive at the end.

 - Do not touch bunker or archive.


---------------------- THIS IS THE VERSION I'M DOING -----------------------------

OTHERWISE we can run in an otherproject mode where we ourselves
hardlink a copy of the BDFs to no_archive and then at the end just
delete all the ones from no_archive - never touching the stuff in the
other bunker or archive directories. This would allow us to test on
other datasets, too, without interruption to anyone else.

For this, the following changes are needed:

 - Add option to mcaf_monitor and to queue_monitor.

 - In mcaf_monitor (?), determine BDF list and hardlink them from bunker.

 - In queue_monitor, don't hardlink to archive, just remove BDFs.

 - Error-check for input tags, i.e. make sure mcaf_mon and queue_mon both have the same option set, and make sure that we're not setting both "archive_all" and "handsoff_archive"

 - Throughout ensure compatibility with test and production versions.

 ---------------------------------------------------------------------------------
"""

import datetime
import os
import asyncore
import click
import logging
import sdmreader 

logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')
logger = logging.getLogger(__name__)

from realfast import queue_monitor, rtutils, mcaf_library

# set up
rtparams_default = os.path.join(os.path.join(os.path.split(os.path.split(mcaf_library.__file__)[0])[0], 'conf'), 'rtpipe_cbe.conf') # install system puts conf files here. used by queue_rtpipe.py
default_bdfdir = '/lustre/evla/wcbe/data/no_archive'
telcaldir = '/home/mchammer/evladata/telcal'  # then yyyy/mm
workdir = os.getcwd()     # assuming we start in workdir
redishost = os.uname()[1]  # assuming we start on redis host


class FRBController(object):
    """Listens for OBS packets and tells FRB processing about any
    notable scans."""

    def __init__(self, intent='', project='', production=False, verbose=False, nrao_controls_archiving=False, rtparams='', slow=0):
        # Mode can be project, intent
        self.intent = intent
        self.project = project
        self.production = production
        self.nrao_controls_archiving = nrao_controls_archiving
        self.verbose = verbose
        self.rtparams = rtparams
        self.slow = slow

    def add_sdminfo(self, sdminfo):
        config = mcaf_library.MCAST_Config(sdminfo=sdminfo)

        # !!! Wrapper here to deal with potential subscans?

        # Check if MCAST message is simply telling us the obs is finished
        if config.obsComplete:
            logger.info("Received finalMessage=True; This observation has completed.")
            # if completing the desired SB, then do a final rsync
            if self.project in config.projectID and self.production:
                logger.info("Final rsync to make workdir copy of SDM %sd complete." % (config.sdmLocation.rstrip('/')))
                rtutils.rsync(config.sdmLocation.rstrip('/'), workdir)  # final transfer to ensure complete SDM in workdir

        # Otherwise check if project is the one we want.
        elif self.project in config.projectID:  # project of interest
            sdmlocation = config.sdmLocation.rstrip('/')
            filename = os.path.join(workdir, os.path.basename(sdmlocation))   # new full-path filename
            scan = int(config.scan)
            bdfloc = os.path.join(default_bdfdir, os.path.basename(config.bdfLocation))

            
            # If we're not controlling the archiving (as would
            # occur if we wanted to make BDFs instantly flowing to
            # the archive, or if we wanted to e.g. test by
            # piggybacking on another (NON-SCIENCE!)
            # observation...
            #
            # We need to then make our own copy in no_archive of
            # the BDF.
            if self.nrao_controls_archiving:
                # Link from bunker to no_archive
                bdfFROM = os.path.join('/lustre/evla/wcbe/data/bunker/',os.path.basename(config.bdfLocation))
                bdfTO   = bdfloc
                if self.production:
                    logger.debug('Hardlinking %s to %s for parallel use by realfast.' % ( bdfFROM, bdfTO ))
                    logger.debug('***********WARNING: QUEUE_MONITOR SHOULD HAVE -N TURNED ON!!!')
                    os.link( bdfFROM, bdfTO )
                else:
                    logger.info('TEST MODE. Would hardlink %s to %s' % ( bdfFROM, bdfTO ))
                    touch( bdfTO + ".test" )

            if self.intent in config.intentString:
                logger.info("Scan %d has desired intent (%s) and project (%s)" % (scan, self.intent, self.project))

                # If we're not in listening mode, prepare data and submit to queue system
                if self.production:
                    assert len(sdmlocation) and isinstance(sdmlocation, str), 'sdmlocation empty or not a string?'
                    assert 'mchammer' not in filename, 'filename %s is SDM original!'

                    # check that SDM is usable by rtpipe. Currently checks spw order and duplicates.
                    if rtutils.check_spw(sdmlocation, scan) and os.path.exists(bdfloc):
                        # 1) copy data into place
                        rtutils.rsync(sdmlocation, workdir)

                        logger.info("Processing sdm %s, scan %d..." % (filename, scan))
                        logger.debug("BDF is in %s\n" % (bdfloc))

                        # 2) find telcalfile (use timeout to wait for it to be written)
                        telcalfile = rtutils.gettelcalfile(telcaldir, filename, timeout=60)

                        # 3) if cal available and bdf exists, submit search job and add tail job to monitoring queue
                        if telcalfile:
                            logger.info('Submitting job to rtutils.search with args: %s %s %s %s %s %s %s %s' % ('default', filename, self.rtparams, '', str([scan]), telcalfile, redishost, default_bdfdir))
                            lastjob = rtutils.search('default', filename, self.rtparams, '', [scan], telcalfile=telcalfile, redishost=redishost, bdfdir=default_bdfdir)
                            rtutils.addjob(lastjob.id)                            
                        else:
                            logger.info('No calibration available. No job submitted.')
                    else:
                        logger.info("Not submitting scan %d of sdm %s. bdf not found or cannot be processed by rtpipe." % (scan, os.path.basename(sdmlocation)))                    

            # each CAL and TARGET scan gets integrated to MS for slow transients search, if in production mode and slow timescale set
            if (self.slow > 0) and ( ('CALIBRATE' in config.intentString) or ('TARGET' in config.intentString) and self.production):
                rtutils.rsync(sdmlocation, workdir)
                logger.info('Creating measurement set for %s scan %d' % (filename, scan))
                sc,sr = sdmreader.read_metadata(filename, scan, bdfdir=default_bdfdir)
                rtutils.linkbdfs(filename, sc, default_bdfdir)
                rtutils.integrate(filename, str(scan), self.slow, redishost)                    
                    
@click.command()
@click.option('--intent', '-i', default='', help='Intent to trigger on')
@click.option('--project', '-p', default='', help='Project name to trigger on')
@click.option('--production', help='Run the code (not just print, etc)', is_flag=True)
@click.option('--verbose', '-v', help='More verbose output', is_flag=True)
@click.option('--nrao_controls_archiving', '-N', help='NRAO controls archiving; i.e. we make our own BDF hardlinks to no_archive and do no archiving at the end. If this is selected, queue_monitor MUST ALSO be run with this same option set.', is_flag=True)
@click.option('--rtparams', help='Parameter file for rtpipe. Default is rtpipe_cbe_conf.', default=rtparams_default)
@click.option('--slow', help='Integrate scans to MS with ints of this many seconds.', default=0)
def monitor(intent, project, production, verbose, nrao_controls_archiving, rtparams, slow):
    """ Monitor of mcaf observation files. 
    Scans that match intent and project are searched (if in production mode).
    Blocking function.
    """

    # Set up verbosity level for log
    if verbose:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)

    logger.info('mcaf_monitor started')
    logger.info('Looking for intent = %s, project = %s' % (intent, project))
    logger.debug('Running in verbose mode')

    if not production:
        logger.info('Running in test mode')
    else:
        logger.info('Running in production mode')

    if nrao_controls_archiving:
        logger.info('*** NRAO CONTROLLING ARCHIVING. ***')
        logger.info('*** Will link and clean BDFs in no_archive directory without ever touching archive directory.')

    # This starts the receiving/handling loop
    controller = FRBController(intent=intent, project=project, production=production, verbose=verbose, nrao_controls_archiving=nrao_controls_archiving, rtparams=rtparams, slow=slow)
    sdminfo_client = mcaf_library.SdminfoClient(controller)
    try:
        asyncore.loop()
    except KeyboardInterrupt:
        # Just exit without the trace barf
        logger.info('Escaping mcaf_monitor')

def testrtpipe(filename, paramfile):
    """ Function for a quick test of rtpipe and queue system
    filename should have full path.
    """

    import sdmreader
    sc,sr = sdmreader.read_metadata(filename, bdfdir=default_bdfdir)
    scan = sc.keys()[0]
    telcalfile = rtutils.gettelcalfile(telcaldir, filename, timeout=60)
    lastjob = rtutils.search('default', filename, paramfile, '', [scan], telcalfile=telcalfile, redishost=redishost, bdfdir=default_bdfdir)
    return lastjob

# Temporary method for creating an empty file.                                                                                                                
def touch(path):
    with open(path, 'a'):
        os.utime(path, None)



""" Functions used in realfast system.
Originally a helper script, so strucutre is odd and needs reworking.
"""
import rtpipe.RT as rt
import rtpipe.calpipe as cp
import rtpipe.parsesdm as ps
import rtpipe.parsecands as pc
from rtpipe import reproduce
import cPickle as pickle
import os, glob, time, shutil, subprocess, logging
import sdmreader

default_bdfdir = '/lustre/evla/wcbe/data/no_archive'
logger = logging.getLogger(__name__)


def search(qname, filename, paramfile, fileroot, scans=[], telcalfile='', redishost='localhost', depends_on=None, bdfdir=default_bdfdir, seggroup=0):
    """ Search for transients in all target scans and segments
    seggroup defines number of segments to send for each pipeline call. default is all segments in one job/node.
    last jobid of group is returned
    """

    from rq import Queue
    from redis import Redis

    # enqueue jobs
    stateseg = []
    logger.info('Setting up pipelines for %s, scans %s...' % (filename, scans))

    for scan in scans:
        assert isinstance(scan, int), 'Scan should be an integer'
        scanind = scans.index(scan)
        state = rt.set_pipeline(filename, scan, paramfile=paramfile, fileroot=fileroot, gainfile=telcalfile, writebdfpkl=True, logfile=False, bdfdir=bdfdir)
        assert state['inttime'] > 0, 'inttime parsed as zero from metadata. Try again?'
        if seggroup:    # submit groups of segments to reduce read/prep overhead
            for segments in grouprange(0, state['nsegments'], seggroup):
                stateseg.append( (state, segments) )
        else:   # all segments in one job/node
            stateseg.append( (state, range(state['nsegments'])) )
    njobs = len(stateseg)

    if njobs:
        logger.info('Enqueuing %d job%s...' % (njobs, 's'[:njobs-1]))

        # submit to queue
        q = Queue(qname, connection=Redis(redishost))

        # enqueue all but one
        if njobs > 1:
            for i in range(njobs-1):
                state, segments = stateseg[i]
                job = q.enqueue_call(func=rt.pipeline, args=(state, segments), depends_on=depends_on, timeout=7*24*3600, result_ttl=7*24*3600)
        else:
            job = depends_on

        # use second to last job as dependency for last job
        state, segments = stateseg[-1]
        lastjob = q.enqueue_call(func=rt.pipeline, args=(state, segments), depends_on=job, timeout=7*24*3600, result_ttl=7*24*3600)  # at_front option makes it hard to predict when jobs will complete

        logger.info('Jobs enqueued. Returning last job with id %s.' % lastjob.id)
        return lastjob
    else:
        logger.info('No jobs to enqueue')
        return


def linkbdfs(filename, scandict=None, bdfdir=default_bdfdir):
    """ Takes proto-sdm filename and makes soft links to create true sdm.
    scandict is optional dictionary from sdmreader that defines scans to link (and the bdf location).
    """

    if not scandict: scandict, sourcedict = sdmreader.read_metadata(filename, bdfdir=bdfdir)

    # Create ASDMBinary directory in our local SDM
    ASDMBinarydir = os.path.join(os.path.basename(filename.rstrip('/')), 'ASDMBinary')
    if not os.path.exists(ASDMBinarydir):
        os.makedirs(ASDMBinarydir)

    # Put BDF softlinks in the ASDMBinary directory to get converted to ms
    for scan in scandict.keys():
        bdfORIG = scandict[scan]['bdfstr'].rstrip('/')
        bdfLINK = os.path.join(ASDMBinarydir, os.path.basename(bdfORIG))
        logger.debug('Creating softlink %s to BDF %s' % (bdfLINK,bdfORIG) )
        if not os.path.exists(bdfLINK):
            os.symlink(bdfORIG, bdfLINK)


def calibrate(filename, fileroot):
    """ Run calibration pipeline
    """

    pipe = cp.pipe(filename, fileroot)
    pipe.run()


def addjob(jobid):
    """ Adds jobid as key in db. Value = 0.
    """

    from redis import Redis
    conn = Redis(db=1)   # db for tracking ids of tail jobs

    conn.set(jobid, 0)


def removejob(jobid):
    """ Removes jobid from db.
    """

    from redis import Redis
    conn = Redis(db=1)   # db for tracking ids of tail jobs

    status = conn.delete(jobid)
    if status:
        logger.info('jobid %s removed from tracking queue' % jobid)
    else:
        logger.info('jobid %s not removed from tracking queue' % jobid)


def mergems(filename, scans, redishost=None, outfile=None):
    import tasklib

    if not outfile: outfile = filename.rstrip('/') + '_slow.ms'

    # first find scans available as ms files
    msscans = []; filelist = []
    for s in scans:
        msfile = filename.rstrip('/') + '_sc' + str(s) + '.ms'
        if os.path.exists(msfile):
            filelist.append(msfile)
            msscans.append(s)
        else:
            logger.debug('Scan %d has no MS' % s)

    if filelist:
        msscanstr= ','.join(str(s) for s in msscans)
        logger.info('Merging slow MS files for scans %s into %s' % (msscanstr, outfile))

        if redishost:
            from rq import Queue
            from redis import Redis

            q = Queue('slow', connection=Redis(redishost))
            q.enqueue_call(func=tasklib.concat, args=(filelist, outfile), timeout=24*3600, result_ttl=24*3600)
        else:
            tasklib.concat(filelist, outfile)
    else:
        logger.info('No ms files found to merge.')


def integrate(filename, scanstr, inttime, redishost=None):
    """ Creates MS from SDM and integrates down.
    filename is sdm, scanstr is comma-delimited string of scans, inttime is time in s (no label).
    filename should be full path, if running job in queue.
    """

    if redishost:
        from rq import Queue
        from redis import Redis

        q = Queue('slow', connection=Redis(redishost))
        q.enqueue_call(func=ps.sdm2ms, args=(filename, filename.rstrip('/') + '_sc' + scanstr + '.ms', scanstr, inttime), timeout=7*24*3600, result_ttl=7*24*3600)
    else:
        ps.sdm2ms(filename, filename.rstrip('/') + '_sc' + scanstr + '.ms', scanstr, inttime)


#### deprecated: moving to rtpipe
def read(filename, paramfile='', fileroot='', bdfdir=default_bdfdir):
    """ Simple parse and return metadata for pipeline for first scan
    """

    sc, sr = sdmreader.read_metadata(filename, bdfdir=bdfdir)
    logger.info('Scans, Target names:')
    logger.info('%s' % str([(ss, sc[ss]['source']) for ss in sc]))
    logger.info('Example pipeline:')
    state = rt.set_pipeline(filename, sc.popitem()[0], paramfile=paramfile, fileroot=fileroot, logfile=False)

def cleanup(workdir, fileroot, scans=[]):
    """ Cleanup up noise and cands files.
    Finds all segments in each scan and merges them into single cand/noise file per scan.
    """

    os.chdir(workdir)

    # merge cands/noise files per scan
    for scan in scans:
        pc.merge_segments(fileroot, scan, cleanup=True, sizelimit=2.)

def merge_scans(workdir, fileroot, scans, snrmin=0, snrmax=999):
    """ Merge cands/noise files over all scans """

    os.chdir(workdir)

    pkllist = [ff for ff in
               ['cands_{0}_sc{1}.pkl'.format(fileroot, scan)
                for scan in scans] if os.path.exists(ff)]
    pc.merge_cands(pkllist, outroot=fileroot, snrmin=snrmin, snrmax=snrmax)

    pkllist = [ff for ff in
               ['noise_{0}_sc{1}.pkl'.format(fileroot, scan) 
                for scan in scans] if os.path.exists(ff)]
    pc.merge_noises(pkllist, fileroot)


def compilenotebook(workdir, fileroot):
    """ Run analysis pipeline from jupyter base notebook and save as notebook and html. """

    os.chdir(workdir)

    os.environ['fileroot'] = fileroot
    cmd = 'jupyter nbconvert baseinteract.ipynb --output {0}.ipynb --to notebook --execute --allow-errors --ExecutePreprocessor.timeout=3600'.format(fileroot).split(
' ')
    status = subprocess.call(cmd)

    cmd = 'jupyter trust {0}.ipynb'.format(fileroot).split(' ')
    status = subprocess.call(cmd)

    cmd = 'jupyter nbconvert {0}.ipynb --to html --output {0}.html'.format(fileroot).split(' ')
    status = subprocess.call(cmd)

def thresholdcands(candsfile, threshold, numberperscan=1):
    """ Returns list of significant candidate loc in candsfile.
    Can define threshold and maximum number of locs per scan.
    Works on merge or per-scan cands pkls.
    """

    # read metadata and define columns of interest
    d = pickle.load(open(candsfile, 'r'))
    try:
        scancol = d['featureind'].index('scan')
    except ValueError:
        scancol = -1
    if 'snr2' in d['features']:
        snrcol = d['features'].index('snr2')
    elif 'snr1' in d['features']:
        snrcol = d['features'].index('snr1')

    # read data and define snrs
    loc, prop = pc.read_candidates(candsfile)
    snrs = [prop[i][snrcol] for i in range(len(prop)) if prop[i][snrcol] > threshold]

    # calculate unique list of locs of interest
    siglocs = [list(loc[i]) for i in range(len(prop)) if prop[i][snrcol] > threshold]
    siglocssort = sorted(zip([list(ll) for ll in siglocs], snrs), key=lambda stuff: stuff[1], reverse=True)

    if scancol >= 0:
        scanset = list(set([siglocs[i][scancol] for i in range(len(siglocs))]))
        candlist= []
        for scan in scanset:
            logger.debug('looking in scan %d' % scan)
            count = 0
            for sigloc in siglocssort:
                if sigloc[0][scancol] == scan:
                    logger.debug('adding sigloc %s' % str(sigloc))
                    candlist.append(sigloc)
                    count += 1
                if count >= numberperscan:
                    break
    else:
        candlist = siglocssort[:numberperscan]

    logger.debug('Returning %d cands above threshold %.1f' % (len(candlist), threshold))
    return [loc for loc,snr in candlist]

#### moving to rtpipe

def plot_cand(candsfile, candloc, redishost=None, **kwargs):
    """ Visualize a candidate as png.
    Can take merge pkl or from a single scan.
    if redishost defined, will submit job to its slow queue, else run locally.
    """

    if redishost:
        from rq import Queue
        from redis import Redis

        q = Queue('slow', connection=Redis(redishost))
        q.enqueue_call(func=reproduce.plot_cand, args=(candsfile, candloc), kwargs=kwargs, timeout=7*24*3600, result_ttl=7*24*3600)
    else:
        reproduce.plot_cand(candsfile, candloc=candloc, **kwargs)


def plot_pulsar(workdir, fileroot, scans=[]):
    """
    Assumes 3 or 4 input pulsar scans (centered then offset pointings).
    """

    pkllist = []
    for scan in scans:
        pkllist.append(os.path.join(workdir, 'cands_' + fileroot + '_sc' + str(scan) + '.pkl'))

    logger.info('Pulsar plotting for pkllist:', pkllist)
    pc.plot_psrrates(pkllist, outname=os.path.join(workdir, 'plot_' + fileroot + '_psrrates.png'))


def getscans(filename, scans='', sources='', intent='', bdfdir=default_bdfdir):
    """ Get scan list as ints.
    First tries to parse scans, then sources, then intent.
    """

    # if no scans defined, set by mode context
    if scans:
        scans = [int(i) for i in scans.split(',')]
    elif sources:
        meta = sdmreader.read_metadata(filename, bdfdir=bdfdir)

        # if source list provided, parse it then append all scans to single list
        sources = [i for i in sources.split(',')]
        scans = []
        for source in sources:
            sclist = filter(lambda sc: source in meta[0][sc]['source'], meta[0].keys())
#            sclist = [sc for sc in meta[0].keys() if source in meta[0][sc]['source']]
            if len(sclist):
                scans += sclist
            else:
                logger.info('No scans found for source %s' % source)
    elif intent:
        meta = sdmreader.read_metadata(filename, bdfdir=bdfdir)
        scans = filter(lambda sc: intent in meta[0][sc]['intent'], meta[0].keys())
#        scans = [sc for sc in meta[0].keys() if intent in meta[0][sc]['intent']]   # get all target fields
    else:
        logger.error('Must provide scans, sources, or intent.')
        raise BaseException

    return scans


def grouprange(start, size, step):
    arr = range(start,start+size)
    return [arr[ss:ss+step] for ss in range(0, len(arr), step)]


def rsync(original, new, mode='-a'):
    """ Uses subprocess.call to rsync from 'filename' to 'new'
    If new is directory, copies original in.
    If new is new file, copies original to that name.
    mode '-a' for archiving (recursively). '-ptgo' better for files or when using wildcard.
    """

    assert os.path.exists(original) or '*' in original, 'original file found or is not a wildcard.'
    assert mode[0] == '-', 'Mode is an option for rsync and must start with \'-\''

    # need to dynamically define whether rsync from has a slash at the end
    if os.path.exists(new):   # new is a directory
        assert os.path.isdir(new), 'New should be a directory.'
        trailing = ''
    else:   # new is a new file. fill it with contents of original
        trailing = '/'

    subprocess.call(["rsync", mode, original.rstrip('/') + trailing, new.rstrip('/')])


def moveplots(fileroot, destination='/users/claw/public_html/realfast'):
    """ For given fileroot, move cand html plot and candidate plots out
    html to destination, candplots to destination/plots
    """

    mergehtml = fileroot + '.html'
    if os.path.exists(mergehtml):
        shutil.copy(mergehtml, destination)
        logger.info('Copied html summary to %s.' % destination)
    else:
        logger.warn('No interactive plot found to copy.')

    candfiles = glob.glob('cands_{0}*.png'.format(fileroot))
    for candfile in candfiles:
        shutil.copy(candfile, os.path.join(destination, 'plots'))
    if candfiles:
        logger.info('Candidate plots copied to %s/plots' % destination)
    else:
        logger.warn('No candidate plots found to copy.')


def copysdm(filename, workdir):
    """ Copies sdm from filename (full path) to workdir
    """

    # first copy data to working area
    fname = os.path.basename(filename)
    newfileloc = os.path.join(workdir, fname)
    if not os.path.exists(newfileloc):
        logger.info('Copying %s into %s' % (fname, workdir))
        shutil.copytree(filename, newfileloc)  # copy file in
    else:
        logger.info('File %s already in %s. Using that one...' % (fname, workdir))
    filename = newfileloc


def copyDirectory(src, dest):
    try:
        shutil.copytree(src, dest)
    # Directories are the same
    except shutil.Error as e:
        print('Directory not copied. Error: %s' % e)
    # Any error saying that the directory doesn't exist
    except OSError as e:
        print('Directory not copied. Error: %s' % e)


def check_spw(sdmfile, scan):
    """ Looks at relative freq of spw and duplicate spw_reffreq. 
    Returns 1 for permutable order with no duplicates and 0 otherwise (i.e., funny data)
    """

    d = rt.set_pipeline(sdmfile, scan, logfile=False)

    dfreq = [d['spw_reffreq'][i+1] - d['spw_reffreq'][i] for i in range(len(d['spw_reffreq'])-1)]
    dfreqneg = [df for df in dfreq if df < 0]

    duplicates = list(set(d['spw_reffreq'])).sort() != d['spw_reffreq'].sort()

    return len(dfreqneg) <= 1 and not duplicates


def find_archivescans(mergefile, threshold=0):
    """ Parses merged cands file and returns list of scans with detections.
    All scans with cand SNR>threshold are returned as a list.
    Goal for this function is to apply RFI rejection, dm-t island detection, and whatever else we can think of.
    """

    # read metadata and define columns of interest
    d = pickle.load(open(mergefile, 'r'))
    scancol = d['featureind'].index('scan')
    if 'snr2' in d['features']:
        snrcol = d['features'].index('snr2')
    elif 'snr1' in d['features']:
        snrcol = d['features'].index('snr1')

    # read data and define snrs
    loc, prop = pc.read_candidates(mergefile)
    snrs = [prop[i][snrcol] for i in range(len(prop))]

    # calculate unique list of scans of interest
    sigscans = [loc[i,scancol] for i in range(len(loc)) if snrs[i] > threshold]
    sigsnrs = [snrs[i] for i in range(len(loc)) if snrs[i] > threshold]
    siglocs = [list(loc[i]) for i in range(len(loc)) if snrs[i] > threshold]
    if len(sigscans):
        logger.info('cands above threshold %.1f for %s: %s' % (threshold, mergefile, str(zip(siglocs, sigsnrs))))
    else:
        logger.info('no cands above threshold %.1f for %s' % (threshold, mergefile))

    return set(sigscans)


def gettelcalfile(telcaldir, filename, timeout=0):
    """ Looks for telcal file with name filename.GN in typical telcal directories
    Searches recent directory first, then tries tree search.
    If none found and timeout=0, returns empty string. Else will block for timeout seconds.
    """

    fname = os.path.basename(filename)
    time_filestart = time.time()

    # search for associated telcal file
    year = str(time.localtime()[0])
    month = '%02d' % time.localtime()[1]
    telcaldir2 = os.path.join(telcaldir, year, month)

    while 1:
        logger.info('Looking for telcalfile in %s' % telcaldir2)
        telcalfile = [os.path.join(telcaldir2, ff) for ff in os.listdir(telcaldir2) if fname+'.GN' in ff]
        
        # if not in latest directory, walk through whole structure
        if not len(telcalfile):
            logger.info('No telcal in newest directory. Searching whole telcalfile tree.')
            telcalfile = [os.path.join(root, fname+'.GN') for root, dirs, files in os.walk(telcaldir) if fname+'.GN' in files]

        assert isinstance(telcalfile, list)

        # make into string (emtpy or otherwise)
        if len(telcalfile) == 1:
            telcalfile = telcalfile[0]
            logger.info('Found telcal file at %s' % telcalfile)
            break
        elif len(telcalfile) > 1:
            telcalfile = ''
            logger.info('Found multiple telcalfiles %s' % telcalfile)
        else:
            telcalfile = ''
            logger.info('No telcal file found in %s' % telcaldir)

        assert isinstance(telcalfile, str)

        # if waiting, but no file found, check timeout
        if timeout:
            if time.time() - time_filestart < timeout:  # don't break yet
                logger.info('Waiting for telcalfile...')
                time.sleep(2)
                continue
            else:   # reached timeout
                logger.info('Timeout waiting for telcalfile')
                break
        else:  # not waiting
            logger.info('Not waiting for telcalfile')
            break

    return telcalfile


def lookforfile(lookdir, subname, changesonly=False):
    """ Look for and return a file with subname in lookdir.
    changesonly means it will wait for changes. default looks only once.
    """

    logger.info('Looking for %s in %s.' % (subname, lookdir))

    filelist0 = os.listdir(os.path.abspath(lookdir))
    if changesonly:
        while 1:
            filelist = os.listdir(os.path.abspath(lookdir))
            newfiles = filter(lambda ff: ff not in filelist0, filelist)
            matchfiles = filter(lambda ff: subname in ff, newfiles)
            if len(matchfiles):
                break
            else:
                logger.info('.')
                filelist0 = filelist
                time.sleep(1)
    else:
        matchfiles = filter(lambda ff: subname in ff, filelist0)

    if len(matchfiles) == 0:
        logger.info('No file found.')
        fullname = ''
    elif len(matchfiles) == 1:
        fullname = os.path.join(lookdir, matchfiles[0])
    elif len(matchfiles) > 1:
        logger.info('More than one match!', matchfiles)
        fullname = os.path.join(lookdir, matchfiles[0])

    logger.info('Returning %s.' % fullname)
    return fullname


def waitforsdm(filename, timeout=300):
    """ Monitors filename (an SDM) to see when it is finished writing.
    timeout is time in seconds to wait from first detection of file.
    Intended for use on CBE.
    """

    time_filestart = 0
    while 1:
        try:
            sc,sr = sdmreader.read_metadata(filename, bdfdir=bdfdir)
        except RuntimeError:
            logger.info('File %s not found.' % filename)
        except IOError:
            logger.info('File %s does not have Antenna.xml yet...' % filename)
            time.sleep(2)
            continue
        else:
            bdflocs = [sc[ss]['bdfstr'] for ss in sc]
            if not time_filestart:
                logger.info('File %s exists. Waiting for it to complete writing.' % filename)
                time_filestart = time.time()
   
     # if any bdfstr are not set, then file not finished
        if None in bdflocs:
            if time.time() - time_filestart < timeout:
                logger.info('bdfs not all written yet. Waiting...')
                time.sleep(2)
                continue
            else:
                logger.info('Timeout exceeded. Exiting...')
                break
        else:
            logger.info('All bdfs written. Continuing.')
            break


def sdmascal(filename, calscans='', bdfdir=default_bdfdir):
    """ Takes incomplete SDM (on CBE) and creates one corrected for use in calibration.
    optional calscans is casa-like string to select scans
    """

    if not calscans:
        calscanlist = getscans(filename, intent='CALI')   # get calibration scans
        calscans = ','.join([str(sc) for sc in calscanlist])   # put into CASA-like selection string

    # make new Main.xml for relevant scans
    if ( os.path.exists(os.path.join(filename, 'Main_cal.xml')) and os.path.exists(os.path.join(filename, 'ASDMBinary_cal')) ):
        logger.info('Found existing cal xml and data. Moving in...')
        shutil.copyfile(os.path.join(filename, 'Main.xml'), os.path.join(filename, 'Main_orig.xml'))   # put new Main.xml in
        shutil.move(os.path.join(filename, 'Main_cal.xml'), os.path.join(filename, 'Main.xml'))   # put new Main.xml in
        shutil.move(os.path.join(filename, 'ASDMBinary_cal'), os.path.join(filename, 'ASDMBinary'))
    else:
        shutil.copyfile(os.path.join(filename, 'Main.xml'), os.path.join(filename, 'Main_orig.xml'))   # put new Main.xml in
        subprocess.call(['choose_SDM_scans.pl', filename, os.path.join(filename, 'Main_cal.xml'), calscans])  #  modify Main.xml
        shutil.move(os.path.join(filename, 'Main_cal.xml'), os.path.join(filename, 'Main.xml'))   # put new Main.xml in

        if not os.path.exists(os.path.join(filename, 'ASDMBinary')):
            os.makedirs(os.path.join(filename, 'ASDMBinary'))

        sc,sr = sdmreader.read_metadata(filename, bdfdir=bdfdir)
        for calscan in calscanlist:
            bdfstr = sc[calscan]['bdfstr'].split('/')[-1]
            bdffile = glob.glob(os.path.join(bdfdir, bdfstr))[0]
            bdffiledest = os.path.join(filename, 'ASDMBinary', os.path.split(bdffile)[1])
            if not os.path.exists(bdffiledest):
                logger.info('Copying bdf in for calscan %d.' % calscan)
                shutil.copyfile(bdffile, bdffiledest)
            else:
                logger.info('bdf in for calscan %d already in place.' % calscan)


def sdmasorig(filename):
    """ Take sdm for calibration and restore it.
    keeps ASDMBinary around, just in case
    """

    shutil.move(os.path.join(filename, 'Main.xml'), os.path.join(filename, 'Main_cal.xml'))
    shutil.move(os.path.join(filename, 'Main_orig.xml'), os.path.join(filename, 'Main.xml'))
    shutil.move(os.path.join(filename, 'ASDMBinary'), os.path.join(filename, 'ASDMBinary_cal'))

